# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
"""For model selector component."""

import json
import os
import shutil
import random
import time
from pathlib import Path
from typing import Union, Callable
from azureml._common._error_definition.azureml_error import AzureMLError
from azureml.acft.common_components.utils.error_handling.error_definitions import InvalidMlflowModelFormat, \
    ModelInputEmpty, ACFTSystemError
from azureml.acft.common_components.utils.error_handling.exceptions import (
    ACFTValidationException,
    ACFTSystemException,
)
from azureml.acft.common_components.utils.logging_utils import get_logger_app
from .constants import (
    ModelSelectorConstants,
    ModelSelectorDefaults,
    ModelSelectorAPIConstants
)
from ..utils.license_utils import save_license_file
from azureml.core.run import Run
import requests
from azureml.dataprep.rslex import Copier, PyIfDestinationExists, PyLocationInfo
from azureml.dataprep.api._rslex_executor import ensure_rslex_environment

logger = get_logger_app(__name__)


class ModelSelector:
    """Model selector class to select the model and store the arguments to json file."""
    def __init__(
        self,
        pytorch_model: str = None,
        mlflow_model: str = None,
        model_name: str = None,
        output_dir: str = None,
        **kwargs,
    ) -> None:
        """Default implementation for model selector. Override the functions for custom implementation.

        :param pytorch_model: asset path of pytorch model, defaults to None
        :type pytorch_model: str, optional
        :param mlflow_model: asset path of mlflow model, defaults to None
        :type mlflow_model: str, optional
        :param model_name: model name from the framework (i.e., HF), defaults to None
        :type model_name: str, optional
        :param output_dir: path to store arguments and model, defaults to None
        :type output_dir: str, optional
        """
        self.pytorch_model = pytorch_model
        self.mlflow_model = mlflow_model
        self.output_dir = output_dir
        self.model_name = model_name
        self.keyword_arguments = kwargs

    def _download_pytorch_model_from_registry(self) -> None:
        """Download pytorch model from AzureML registry"""
        model_path = os.path.join(
            self.output_dir, ModelSelectorDefaults.PYTORCH_MODEL_DIRECTORY
        )
        ModelSelector._copy_directories(self.pytorch_model, model_path)
        logger.info(
            f"Downloaded pytorch model from {self.pytorch_model} to {model_path}."
        )

    def _download_mlflow_model_from_registry(self) -> None:
        """Download mlflow model from AzureML registry"""
        model_path = os.path.join(
            self.output_dir, ModelSelectorDefaults.MLFLOW_MODEL_DIRECTORY
        )
        ModelSelector.convert_mlflow_model_to_pytorch_model(self.mlflow_model, Path(model_path))
        logger.info(
            f"Downloaded mlflow model from {self.mlflow_model} to {model_path}."
        )

    def _prepare_and_logs_arguments(self) -> None:
        """Update the keyword arguments (if present) with required key-val items and
        Store the model selector arguments to json file.
        """
        arguments = {
            ModelSelectorConstants.MLFLOW_MODEL_PATH: self.mlflow_model,
            ModelSelectorConstants.PYTORCH_MODEL_PATH: self.pytorch_model,
            ModelSelectorConstants.MODEL_NAME: self.model_name,
        }

        if self.keyword_arguments:
            self.keyword_arguments.update(arguments)
        else:
            self.keyword_arguments = arguments

        os.makedirs(self.output_dir, exist_ok=True)
        model_selector_args_save_path = os.path.join(
            self.output_dir, ModelSelectorDefaults.MODEL_SELECTOR_ARGS_SAVE_PATH
        )
        logger.info(
            f"Saving the model selector args to {model_selector_args_save_path}"
        )
        with open(model_selector_args_save_path, "w") as output_file:
            json.dump(self.keyword_arguments, output_file, indent=2)

    @staticmethod
    def _copy_directories(asset_path: str, destination: str) -> None:
        """Recursively copy asset to destination

        :param asset_path: asset path of Azure ML
        :type asset_path: str
        :param destination: destination path
        :type destination: str
        """
        os.makedirs(destination, exist_ok=True)
        shutil.copytree(asset_path, destination, dirs_exist_ok=True)

    def convert_mlflow_model_to_pytorch_model(
            mlflow_model_path: Union[str, Path],
            download_dir: Path) -> None:
        """
        Download the mlflow model assets(model, config and proprocessor.json file)
        in the download directory to have similar directory structure as the pytorch model.

        :param mlflow_model_path: Path of the mlflow model
        :type mlflow_model_path: Union[str, Path]
        :param download_dir: Destination directory to download the model
        :type download_dir: Path
        """

        os.makedirs(download_dir, exist_ok=True)
        try:
            DATA_DIR = ModelSelectorConstants.MLFLOW_MODEL_DATA_PATH
            model_dir = Path(mlflow_model_path, f'{DATA_DIR}/model')
            config_dir = Path(mlflow_model_path, f'{DATA_DIR}/config')
            tokenizer_dir = Path(mlflow_model_path, f'{DATA_DIR}/tokenizer')

            # copy the model
            logger.info("downloading mlflow model")
            shutil.copytree(model_dir, download_dir, dirs_exist_ok=True)

            # copy the config files
            logger.info("downloading mlflow config files")
            shutil.copytree(config_dir, download_dir, dirs_exist_ok=True)

            # copy tokenizer files
            logger.info("downloading tokenizer files")
            shutil.copytree(tokenizer_dir, download_dir, dirs_exist_ok=True)

            # copy license file
            save_license_file(
                model_name_or_path=mlflow_model_path,
                license_file_name=ModelSelectorDefaults.LICENSE_FILE_NAME,
                destination_paths=[str(download_dir)]
            )
        except Exception:
            shutil.rmtree(download_dir, ignore_errors=True)
            directories = f" '{model_dir}', '{config_dir}', '{tokenizer_dir}' "
            raise ACFTValidationException._with_error(
                AzureMLError.create(
                    InvalidMlflowModelFormat, directories
                )
            )

    def _call_api_with_retries(self, request_method: Callable[..., requests.Response],
                               uri: str, **kwargs: dict) -> requests.Response:
        """
        Submit a REST API request to the given uri using the given request method. Retries failed requests
        and either fails with error reporting or returns the request reponse.

        :param request_method: request method to call
        :type request_method: Callable[..., requests.Response]
        :param uri: REST API uri
        :type uri: str
        :param **kwargs: keyword args to pass to request method
        :type **kwargs: dict
        """

        for i in range(ModelSelectorAPIConstants.API_RETRY_COUNT - 1):
            try:
                return request_method(uri, **kwargs)
            except requests.exceptions.RequestException as e:
                logger.warning(
                    f"Encountered {e.__class__.__name__} error while calling REST API "
                    f"Retries left: {ModelSelectorAPIConstants.API_RETRY_COUNT - i - 1}. Error details: {e}"
                )

                wait_time = ((2 ** i) * ModelSelectorAPIConstants.BACKOFF_IN_SECONDS + random.uniform(0, 1))
                logger.info(f"Waiting {wait_time:.2f} seconds before attempting call to {uri} again.")
                time.sleep(wait_time)

        # Attempt to load the call REST API one more time. Warn and advise to run again in case of new failure.
        logger.warning("Attempting to call REST API one last time. If this fails, please submit a new run.")
        try:
            return request_method(uri, **kwargs)
        except requests.exceptions.RequestException as e:
            error_string = (
                f"Encountered {e.__class__.__name__} error while calling REST API "
                f"Failed with exception {e}"
            )
            raise ACFTSystemException._with_error(
                AzureMLError.create(ACFTSystemError, pii_safe_message=error_string)
            )

    def _get_model_info_from_registry(self) -> requests.Response:
        """
        Retrieve model metadata from registry using REST API call.
        :return API call response
        :type requests.Response
        """
        if self.service_endpoint is None:
            logger.info(f"{ModelSelectorAPIConstants.SERVICE_ENDPOINT} environment variable is not set")
            return
        asset_id = ModelSelectorAPIConstants.MODEL_ASSET_ID.format(ModelSelectorDefaults.MODEL_REGISTRY,
                                                                   self.registry_model_name)
        registry_uri = ModelSelectorAPIConstants.REGISTRY_URI.format(self.service_endpoint, asset_id)

        # make API call to get model info from registry
        logger.info(f"Calling REST API to get model metadata for {self.registry_model_name} from registry")
        request_args = {"headers" : self.auth_header}
        return self._call_api_with_retries(requests.get, registry_uri, **request_args)

    def _get_blobstore_sas_credential(self, model_info: dict) -> requests.Response:
        """
        Retrieve SaS credential for blob url specified in model_info using REST API call.
        :param model_info: model metadata including blob url
        :type model_info: dict
        :return API call response
        :type requests.Response
        """
        model_url = model_info[ModelSelectorAPIConstants.URL]
        asset_id = model_info[ModelSelectorAPIConstants.ASSET_ID]

        # get SaS credential because we cannot access blob store with AML Run Token
        sas_uri = ModelSelectorAPIConstants.SAS_URI.format(self.service_endpoint)
        blob_reference_sas_request_dto = {
            ModelSelectorAPIConstants.BLOB_REF_ASSET_ID: asset_id,
            ModelSelectorAPIConstants.BLOB_REF_BLOB_URI: model_url
        }

        # make API call to get SaS credential from registry
        logger.info("Calling REST API to get SaS credential for model artifacts blob")
        request_args = {"json" : blob_reference_sas_request_dto, "headers" : self.auth_header}
        return self._call_api_with_retries(requests.post, sas_uri, **request_args)

    def _download_blobstore_model_artifacts(self, sas_info: dict) -> os.PathLike:
        """
        Download model artifacts from blobstore with SaS credential.
        :param sas_info: SaS credential
        :type model_info: dict
        :return model_download_path: path that model artifacts are downloaded to
        :type os.Path
        """
        model_download_path = os.path.join(self.output_dir, ModelSelectorConstants.REGISTRY_DOWNLOAD_DIR)
        os.makedirs(model_download_path, exist_ok=True)
        download_location = PyLocationInfo('Local', model_download_path, {})
        if_exists = PyIfDestinationExists.MERGE_WITH_OVERWRITE
        credential = sas_info[ModelSelectorAPIConstants.BLOB_REF][ModelSelectorAPIConstants.CREDENTIAL]
        wasbs_uri = credential[ModelSelectorAPIConstants.WASBS_URI]
        try:
            logger.info("Downloading model artifacts from blobstore")
            ensure_rslex_environment()
            Copier.copy_uri(download_location, wasbs_uri, if_exists, "")
        except Exception as e:
            error_string = (
                f"Unable to download model artifacts from blobstore."
                f"Failed with exception {e}"
            )
            raise ACFTSystemException._with_error(
                AzureMLError.create(ACFTSystemError, pii_safe_message=error_string)
            )

        logger.info(f"Model {self.registry_model_name} is downloaded to folder {model_download_path}")
        return model_download_path

    def _model_name_download_from_registry(self) -> None:
        """
        Search for model name in ModelSelectorDefaults.MODEL_REGISTRY.
        Download model and set model asset path (self.pytorch_model or self.mlflow_model)
        if the model is successfully downloaded from the registry.
        """

        # convert HF id to registry model name
        self.registry_model_name = self.model_name.replace("/", "-")
        logger.info(f"Searching for model {self.registry_model_name} in registry")

        # get AML run token and service endpoint
        context = Run.get_context()
        self.auth_header = context.experiment.workspace._auth_object.get_authentication_header()
        self.service_endpoint = os.getenv(ModelSelectorAPIConstants.SERVICE_ENDPOINT)

        model_info_response = self._get_model_info_from_registry()

        # if the model was found, get sas credential and download model artifacts
        if model_info_response.status_code == 200:
            logger_str = f"Model {self.registry_model_name} was found in registry "\
                         f"{ModelSelectorDefaults.MODEL_REGISTRY}"
            logger.info(logger_str)
            model_info = model_info_response.json()

            # get SaS credential for accessing blobstore
            sas_info_response = self._get_blobstore_sas_credential(model_info)
            if sas_info_response.status_code != 200:
                error_string = "Unable to retrieve SaS token for model download from registry"
                raise ACFTSystemException._with_error(
                    AzureMLError.create(ACFTSystemError, pii_safe_message=error_string)
                )
            sas_info = sas_info_response.json()
            logger.info("SaS token retrieved to download model")

            # download model artifacts from blobstore using sas credential
            model_download_path = self._download_blobstore_model_artifacts(sas_info)

            # set the asset path for the downloaded model, replace model_name with registry_model_name
            model_format = model_info[ModelSelectorAPIConstants.MODEL_FORMAT]
            if model_format == ModelSelectorAPIConstants.CUSTOM_MODEL:
                self.pytorch_model = Path(model_download_path, ModelSelectorConstants.PYTORCH_MODEL_ROOT)
                self.model_name = self.registry_model_name
            elif model_format == ModelSelectorAPIConstants.MLFLOW_MODEL:
                self.mlflow_model = Path(model_download_path, ModelSelectorConstants.MLFLOW_MODEL_ROOT)
                self.model_name = self.registry_model_name
            else:
                logger.warning(f'Model {self.registry_model_name} is not \
                            of type {ModelSelectorAPIConstants.CUSTOM_MODEL} or \
                            {ModelSelectorAPIConstants.MLFLOW_MODEL}')

        else:
            logger.warning(f'Model {self.model_name} was not found in registry {ModelSelectorDefaults.MODEL_REGISTRY}')

    def run_workflow(self) -> None:
        """If model asset path is provided then it will download the model. Pytorch model will take preference
        over mlflow model. If model name is provided then, it will download the model from the registry. If the
        model is not found in the registry, it will download model from framework hub.
        It's responsibility of downstream component (e.g., finetune) to load the model.

        :raises ArgumentException._with_error: Raise exception if model ports or model name is not provided.
        """

        # If model_name is provided and asset paths are not specified, download the model from the registry.
        # If the model is found in the registry, the asset path will be set for the next if/elif block
        if (self.model_name is not None) and (self.pytorch_model is None) and (self.mlflow_model is None):
            self._model_name_download_from_registry()

        if self.pytorch_model is not None:
            self._download_pytorch_model_from_registry()
            self.pytorch_model = ModelSelectorDefaults.PYTORCH_MODEL_DIRECTORY
            self.mlflow_model = None
        elif self.mlflow_model is not None:
            self._download_mlflow_model_from_registry()
            self.mlflow_model = ModelSelectorDefaults.MLFLOW_MODEL_DIRECTORY
        elif self.model_name is None:
            raise ACFTValidationException._with_error(
                AzureMLError.create(
                    ModelInputEmpty, argument_name="Model ports and model_name"
                )
            )
        self._prepare_and_logs_arguments()
